/*
 * Copyright (c) 2017 Adrian Siekierka
 *
 * This file is part of Unlimited Chisel Works.
 *
 * Unlimited Chisel Works is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Unlimited Chisel Works is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Unlimited Chisel Works.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.ucw;

import net.minecraft.block.Block;
import net.minecraft.block.SoundType;
import net.minecraft.block.material.MapColor;
import net.minecraft.block.properties.IProperty;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.util.BlockRenderLayer;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.NonNullList;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;
import net.minecraftforge.common.property.ExtendedBlockState;
import net.minecraftforge.common.property.IUnlistedProperty;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import javax.annotation.Nullable;
import java.util.Collection;
import java.util.List;

public class BlockUCWProxy extends Block implements IUCWBlock {
	private UCWBlockRule rule;
	private IBlockState base;

	public BlockUCWProxy() {
		super(UCWObjectBroker.get().getBase().func_185904_a());

		rule = UCWObjectBroker.get().getRule();
		base = UCWObjectBroker.get().getBase();

		UnlimitedChiselWorks.proxy.initBlock(base, this);
		func_149663_c(base.func_177230_c().func_149739_a());
	}

	@Override
	public boolean func_149662_c(IBlockState state) {
		return UCWUtils.applyProperties(rule.throughBlock, state).func_185914_p();
	}

	@Override
	public boolean func_149686_d(IBlockState state) {
		return UCWUtils.applyProperties(rule.throughBlock, state).func_185917_h();
	}

	@Override
	public boolean func_149730_j(IBlockState state) {
		return UCWUtils.applyProperties(rule.throughBlock, state).func_185913_b();
	}

	@Override
	public boolean func_149721_r(IBlockState state) {
		return UCWUtils.applyProperties(rule.throughBlock, state).func_185915_l();
	}

	@Override
	public int getLightOpacity(IBlockState state, IBlockAccess world, BlockPos pos) {
		return base.getLightOpacity(new UCWBlockAccess(world), pos);
	}

	@Override
	public int getLightValue(IBlockState state, IBlockAccess world, BlockPos pos) {
		return base.getLightValue(new UCWBlockAccess(world), pos);
	}

	@Override
	@SideOnly(Side.CLIENT)
	public boolean canRenderInLayer(IBlockState state, BlockRenderLayer layer) {
		IBlockState newState = UCWUtils.applyProperties(rule.throughBlock, state);
		return newState.func_177230_c().canRenderInLayer(newState, layer);
	}

	@Override
	@SideOnly(Side.CLIENT)
	public BlockRenderLayer func_180664_k() {
		return rule.throughBlock.func_180664_k();
	}

	@Override
	public boolean func_176200_f(IBlockAccess worldIn, BlockPos pos) {
		return base.func_177230_c().func_176200_f(new UCWBlockAccess(worldIn), pos);
	}

	@Override
	public boolean canHarvestBlock(IBlockAccess world, BlockPos pos, EntityPlayer player) {
		return rule.fromBlock.canHarvestBlock(new UCWBlockAccess(world), pos, player);
	}

	@Override
	public float func_149638_a(Entity exploder) {
		return base.func_177230_c().func_149638_a(exploder);
	}

	@Override
	public SoundType func_185467_w() {
		return base.func_177230_c().func_185467_w();
	}

	@Override
	public boolean isLeaves(IBlockState state, IBlockAccess world, BlockPos pos) {
		return base.func_177230_c().isLeaves(base, new UCWBlockAccess(world), pos);
	}

	@Override
	public boolean isWood(IBlockAccess world, BlockPos pos) {
		return base.func_177230_c().isWood(new UCWBlockAccess(world), pos);
	}

	@Override
	public boolean isFoliage(IBlockAccess world, BlockPos pos) {
		return base.func_177230_c().isFoliage(new UCWBlockAccess(world), pos);
	}

	@Override
	public float func_176195_g(IBlockState blockState, World worldIn, BlockPos pos) {
		try {
			return rule.fromBlock.func_176195_g(base, worldIn, pos);
		} catch (Exception e) {
			try {
				return rule.throughBlock.func_176195_g(UCWUtils.applyProperties(rule.throughBlock, blockState), worldIn, pos);
			} catch (Exception ee) {
				return field_149782_v;
			}
		}
	}

	@Override
	public float func_180647_a(IBlockState state, EntityPlayer player, World worldIn, BlockPos pos) {
		try {
			return rule.fromBlock.func_180647_a(base, player, worldIn, pos);
		} catch (Exception e) {
			try {
				return rule.throughBlock.func_180647_a(UCWUtils.applyProperties(rule.throughBlock, state), player, worldIn, pos);
			} catch (Exception ee) {
				return field_149782_v;
			}
		}
	}

	@Override
	public int getFlammability(IBlockAccess world, BlockPos pos, EnumFacing face) {
		return rule.fromBlock.getFlammability(new UCWBlockAccess(world), pos, face);
	}

	@Override
	public boolean isFlammable(IBlockAccess world, BlockPos pos, EnumFacing face) {
		return rule.fromBlock.isFlammable(new UCWBlockAccess(world), pos, face);
	}

	@Override
	public int getFireSpreadSpeed(IBlockAccess world, BlockPos pos, EnumFacing face) {
		return rule.fromBlock.getFireSpreadSpeed(new UCWBlockAccess(world), pos, face);
	}

	@Override
	public boolean isBeaconBase(IBlockAccess worldObj, BlockPos pos, BlockPos beacon) {
		return rule.fromBlock.isBeaconBase(new UCWBlockAccess(worldObj), pos, beacon);
	}

	@Override
	public int getHarvestLevel(IBlockState state) {
		return base.func_177230_c().getHarvestLevel(base);
	}

	@Override
	public boolean isToolEffective(String type, IBlockState state) {
		return base.func_177230_c().isToolEffective(type, base);
	}

	@Override
	public int func_180651_a(IBlockState state) {
		return func_176201_c(state);
	}

	@Override
	public IBlockState func_176203_a(int meta) {
		return UCWUtils.applyProperties(this, rule.throughBlock.func_176203_a(meta));
	}

	@Override
	public int func_176201_c(IBlockState state) {
		return rule.throughBlock.func_176201_c(UCWUtils.applyProperties(rule.throughBlock, state));
	}

	@Override
	public void func_149666_a(Item itemIn, CreativeTabs tab, NonNullList<ItemStack> items) {
		Item.func_150898_a(this).func_150895_a(Item.func_150898_a(this), tab, items);
	}

	@Override
	protected BlockStateContainer func_180661_e() {
		rule = UCWObjectBroker.get().getRule();
		base = UCWObjectBroker.get().getBase();

		return UCWObjectBroker.get().createBlockState(this);
	}

	@Override
	public IBlockState getBaseState() {
		return base;
	}
}
